@@ -1,5 +1,5 @@
 """
-Copyright 2021 RICHARD TJÖRNHAMMAR
+Copyright 2022 RICHARD TJÖRNHAMMAR
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
@@ -32,8 +32,8 @@ class Node ( object ) :
         self.descendants_ :list  = list() # OUTWARD LINKS , DIRECT DESCENDENCY ( 1 LEVEL )
         self.data_        :dict  = dict() # OTHER THINGS SHOULD BE ALL INFORMATION FLOATING IN USERSPACE
 
-    def can_it_be_root(self) -> bool :
-        return ( len(self.ascendants) == 0 )
+    def can_it_be_root ( self, n:int=1 ) -> bool :
+        return ( len(self.ascendants_) < n )
 
     def supplement ( self, n:super ) -> None :
         self.label_       = n.label_
@@ -49,16 +49,20 @@ class Node ( object ) :
                            links : type(list(str())) ,
                            label : str = "" ,
                            description : str = "" ) -> object :
+        # ASSIGNS ALL META DATA AND BIPOLAR LINKS
         self.set_id( identification )
         self.add_label( label )
         self.add_description( description )
         self.add_links( links , bClear=True )
         return ( self )
 
-    def set_level ( self,level ) :
+    def set_level ( self,level:int ) -> None :
         self.level_ = level
 
-    def level ( self ) :
+    def set_metrics ( self , metrics:list ) -> None :
+        self.metrics_ = [ *self.metrics_ , *metrics ]
+
+    def level ( self ) -> None :
         return ( self.level_ )
 
     def get_data ( self ) -> dict :
@@ -121,6 +125,8 @@ class Node ( object ) :
             s_inf += '\n['+linktype+'] : '
             for l in self.get_links(linktype=linktype) :
                 s_inf += l + '\t'
+        for item in self.get_data().items() :
+            s_inf += '\n'+str(item[0])+'\t'+str(item[1])
         print ( s_inf )
 
 class NodeGraph ( Node ) :
@@ -141,6 +147,13 @@ class NodeGraph ( Node ) :
     def items ( self )  -> list :
         return( self.graph_map_.items() )
 
+    def list_roots ( self ) ->  type(list(str())) :
+        roots = [] # BLOODY ROOTS
+        for name,node in self.items():
+            if node.can_it_be_root() :
+                roots.append( name )
+        return ( roots )
+
     def get_node ( self, nid : str ) -> Node :
         return ( self.graph_map_[nid] )
 
@@ -158,7 +171,10 @@ class NodeGraph ( Node ) :
             if len ( self.graph_map_ ) == 1 :
                 self.set_root_id( n.identification() )
 
-    def get_dag ( self ) -> dict() :
+    def get_dag ( self ) -> dict :
+        return ( self.graph_map_ )
+
+    def get_graph ( self ) -> dict :
         return ( self.graph_map_ )
 
     def show ( self ) -> None :
@@ -176,10 +192,9 @@ class NodeGraph ( Node ) :
         results['path'] = [ idx for idx in results['path'] if not idx==identification ]
         return ( results )
 
-
     def search ( self , order:str = 'breadth', root_id:str = None ,
                  linktype:str = 'links', stop_at:str = None ) -> dict :
-
+        #
         path:list   = list()
         visited:set = set()
         if root_id is None :
@@ -226,6 +241,178 @@ class NodeGraph ( Node ) :
 
         return ( { 'path':path , 'order':order , 'linktype':linktype } )
 
+    def connectivity ( self, distm:np.array , alpha:float , n_connections:int=1 , bOld:bool=True ) -> list :
+        #
+        # AN ALTERNATIVE METHOD
+        # DOES THE SAME THING AS THE CONNECTIVITY CODE IN MY
+        # CLUSTERING MODULE (in src/impetuous/clustering.py )
+        # OR IN https://github.com/richardtjornhammar/RichTools/blob/master/src/cluster.cc
+        # ADDED TO RICHTOOLS HERE: https://github.com/richardtjornhammar/RichTools/commit/74b35df9c623bf03570707a24eafe828f461ed90#diff-25a6634263c1b1f6fc4697a04e2b9904ea4b042a89af59dc93ec1f5d44848a26
+        # CONNECTIVITY SEARCH FOR (connectivity) CONNECTIVITY
+        #
+        # THIS ROUTINE RETURNS A LIST BELONGING TO THE CLUSTERS
+        # WITH THE SET OF INDICES THAT MAPS TO THE CLUSTER
+        # THIS METHOD IS NOW ALSO IN THE clustering.py MODULE
+        # AND IS CALLED connectedness
+        # THIS CLASS WILL EMPLOY THE JIT connectivity IMPLEMENTATION
+        # IN THE FUTURE BECAUSE IT IS SUPERIOR
+        #
+        if len ( distm.shape ) < 2 :
+            print ( 'PLEASE SUBMIT A SQUARE DISTANCE MATRIX' )
+            exit(1)
+        #
+        if bOld : # WATER CLUSTERING ALGO FROM 2009
+            from impetuous.clustering import connectivity as connections
+            results = connections ( distm , alpha )
+            L = [set() for i in range(len(results[0]))]
+            for c in results[1]:
+                L[c[0]] = L[c[0]]|set([c[1]])
+            return ( L )
+        #
+        def b2i ( a:list ) -> list :
+            return ( [ i for b,i in zip(a,range(len(a))) if b ] )
+        def f2i ( a:list,alf:float ) -> list :
+            return ( b2i( a<=alf ) )
+        L = []
+        for a in distm :
+            bAdd = True
+            ids = set( f2i(a,alpha) )
+            for i in range(len(L)) :
+                if len( L[i]&ids ) >=  n_connections :
+                    L[i] = L[i] | ids
+                    bAdd = False
+                    break
+            if bAdd and len(ids) >= n_connections :
+                L .append( ids )
+        return ( L )
+
+    def linkages_to_pclist ( self , links:dict ) -> list :
+        bottoms_up = sorted([ (v,k) for k,v in links.items()])
+        PClist = []
+        while ( len(bottoms_up)>1 ) :
+            child = bottoms_up[0]
+            for parent in bottoms_up[1:] :
+                if child[1] in parent[1] :
+                    parent_child = [ (parent[1], child[1]) ]
+                    PClist = [ *PClist, *parent_child ]
+                    bottoms_up.remove( child )
+                    break
+        return ( PClist )
+
+    def linkages_to_graph_dag( self, links:dict ) -> None :
+        PClist = self.linkages_to_pclist ( links )
+        for pc in PClist :
+            self.add_ascendant_descendant ( pc[0], pc[1] )
+        self.set_root_id(self.list_roots()[0] )
+
+    def distance_matrix_to_pclist ( self , distm:np.array ,
+                                    cluster_connections:int = 1 ,
+                                    hierarchy_connections:int = 1 ,
+                                    bNonRedundant:bool = True  ) -> list :
+        #
+        # FASTER PCLIST CONSTRUCTION ROUTINE
+        # RETURNS LIST USEFUL FOR HIERARCHY GENERATION
+        # SHOULD BE EASIER TO PARALLELIZE WITH JIT
+        #
+        logic = lambda p,c : len(p&c) >= hierarchy_connections and len(p^c)>0
+        if not bNonRedundant :
+            logic = lambda p,c : len(p&c) >= hierarchy_connections
+        #
+        R = sorted( list(set( distm.reshape(-1) ) ) )
+        prev_clusters = []
+        PClist = []
+        for r in R :
+            present_clusters = self.connectivity ( distm , r , cluster_connections )
+            parent_child  = [ (p,c,r) for c in prev_clusters for p in present_clusters \
+                          if logic(p,c)  ]
+            prev_clusters = present_clusters
+            PClist = [ *PClist, *parent_child ]
+        return ( PClist )
+
+    def distance_matrix_to_absolute_coordinates ( self , D:np.array , bSquared:bool = False, n_dimensions:int=2 ) -> np.array :
+        #
+        # SAME AS IN THE IMPETUOUS cluster.py EXCEPT THE RETURN IS TRANSPOSED
+        # AND distg.m IN THE RICHTOOLS REPO
+        # C++ VERSION HERE https://github.com/richardtjornhammar/RichTools/commit/be0c4dfa8f61915b0701561e39ca906a9a2e0bae
+        #
+        if not bSquared :
+            D = D**2.
+        DIM = n_dimensions
+        DIJ = D*0.
+        M = len(D)
+        for i in range(M) :
+            for j in range(M) :
+                DIJ[i,j] = 0.5* (D[i,-1]+D[j,-1]-D[i,j])
+        D = DIJ
+        U,S,Vt = np.linalg.svd ( D , full_matrices = True )
+        S[DIM:] *= 0.
+        Z = np.diag(S**0.5)[:,:DIM]
+        xr = np.dot( Z.T,Vt )
+        return ( xr.T )
+
+    def distance_matrix_to_graph_dag ( self , distm:np.array , n_:int=1 , bVerbose:bool=False , names:list=None ) -> None :
+        #
+        # CONSTRUCTS THE HIERACHY FROM A DISTANCE MATRIX
+        # SIMILAR TO THE ROUTINES IN hierarchical.py IN THIS IMPETUOUS REPO
+        #
+        if len ( distm.shape ) < 2 :
+            print ( 'PLEASE SUBMIT A SQUARE DISTANCE MATRIX' )
+            exit(1)
+        lookup = dict()
+        m_ = len(distm)
+        for I in range(m_) :
+            lookup[I] = I
+        if not names is None :
+            if len ( names ) == m_ :
+                for I,N in zip(range(len(names)),names):
+                    lookup[I] = N
+        pclist = self.distance_matrix_to_pclist( distm )
+        for pc_ in pclist :
+            lpc0 = [ lookup[l] for l in list(pc_[0]) ]
+            lpc1 = [ lookup[l] for l in list(pc_[1]) ]
+            asc = str(lpc0)
+            des = str(lpc1)
+            asc_met = pc_[2]
+            self.add_ascendant_descendant(asc,des)
+            self.get_graph()[asc].set_metrics([asc_met])
+            self.get_graph()[asc].get_data()['analyte ids'] = lpc0
+            self.get_graph()[des].get_data()['analyte ids'] = lpc1
+        for key in self.keys() :
+            if self.get_graph()[key].can_it_be_root(n_):
+                self.set_root_id ( key )
+        if bVerbose :
+            self.show()
+            print ( self.get_root_id() )
+            self.get_graph()[self.get_root_id()].show()
+
+    def graph_analytes_to_approximate_distance_matrix ( self ,
+             analyte_identifier:str = 'analyte ids',
+             alpha:float = 1. ) -> np.array :
+
+        root_data    = self.get_graph()[ self.get_root_id() ].get_data()
+
+        if analyte_identifier in root_data:
+            all_analytes = root_data[ analyte_identifier ]
+        else:
+            print ( 'ERROR COULD NOT FIND GLOBAL IDENTIFIER INFORMATION:' , analyte_identifier )
+            exit (1)
+        m_ = len( all_analytes )
+        lookup = { a:i for a,i in zip(all_analytes,range(m_)) }
+        CM = np.ones(m_*m_).reshape(m_,m_)
+        for item in self.get_graph().items() :
+
+            item_data = item[1].get_data()
+
+            if analyte_identifier in item_data : # IMPROVE SPEED HERE
+                for q in item_data[analyte_identifier] :
+                    for p in item_data[analyte_identifier] : # STOP
+                        CM[lookup[p],lookup[q]]+=1
+        #
+        # CONSTRUCT APPROXIMATION
+        # LEVELS MISSING EVEN VIA DEFAULT
+        approximate_distm  = 1./CM - np.mean(np.diag(1./CM))
+        approximate_distm *= 1-np.eye(m_)
+        return ( np.abs(approximate_distm) , lookup )
 
     def add_ascendant_descendant ( self, ascendant:str, descendant:str ) -> None :
         n = Node()
@@ -243,14 +430,12 @@ class NodeGraph ( Node ) :
         self.add(n)
         self.add(m)
 
-
     def generate_ascendants_descendants_lookup ( self ) -> (type(list(str())),type(list(str()))) :
         all_names   = self.keys()
         descendants = [ ( idx , set( self.complete_lineage( idx,linktype='descendants')['path'] ) ) for idx in all_names ]
         ancestors   = [ ( idx , set( self.complete_lineage( idx,linktype='ascendants' )['path'] ) ) for idx in all_names ]
         return ( ancestors , descendants )
 
-
     def ascendant_descendant_file_to_dag ( self, relationship_file:str = './PCLIST.txt' ,
                                   i_a:int = 0 , i_d:int = 1 ,
                                   identifier:str = None , sep:str = '\t' ) -> (type(list(str())),type(list(str()))) :
@@ -384,9 +569,9 @@ def make_pathway_ancestor_data_frame(ancestors):
             p_df = pd.concat([p_df,t_df])
     return( p_df )
 
-def normalise_for_apples_and_oranges_stats( X , method='ordinal' ):
-    X_ = rankdata( X , method=method )/len(X)
-    return(X_)
+def normalise_for_apples_and_oranges_stats( X:np.array , method:str='average' ) -> np.array :
+    X_ = (rankdata( X , method=method )-0.5)/len(set(X))
+    return ( X_ )
 
 def make_group_analytes_unique( grouping_file , delimiter='\t' ):
     uniqe_grouping_file = '/'.join(grouping_file.split('/')[:-1]) + '/unique_'+grouping_file.split('/')[-1]
@@ -500,12 +685,15 @@ def convert_rdata_to_dataframe ( filename ) :
     pandas2ri.deactivate()
     return ( full_df_dict )
 
-def read_xyz(fname='argon.xyz') :
-    df_ = pd.read_csv('./argon.xyz',header=2,sep=' ')
-    vals = df_.columns.values
-    df_.loc[len(df_)] = vals
-    df_.columns = ['A','X','Y','Z']
-    return ( df_.apply(pd.to_numeric) )
+def read_xyz(fname='argon.xyz',sep=' ') :
+    coords = []
+    with open(fname,'r') as input:
+        for line in input:
+            lsp = [u for u in line.replace('\n','').split(sep) if len(u)>0 ]
+            print(lsp,len(lsp))
+            if len(lsp) == 4:
+                coords.append( ( lsp[0],[ float(c) for c in lsp[1:]] ) )
+    return ( coords )
 
 
 import os
