@@ -1,5 +1,5 @@
 """
-Copyright 2021 RICHARD TJÖRNHAMMAR
+Copyright 2022 RICHARD TJÖRNHAMMAR
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
@@ -21,17 +21,18 @@ from scipy.stats import ttest_rel , ttest_ind , mannwhitneyu
 from scipy.stats.mstats import kruskalwallis as kruskwall
 from sklearn.decomposition import PCA
 import itertools
+import typing
 
-def SubArraysOf ( Array,Array_=None ) :
+def subArraysOf ( Array:list,Array_:list=None ) -> list :
     if Array_ == None :
         Array_ = Array[:-1]
     if Array == [] :
         if Array_ == [] :
             return ( [] )
-        return( SubArraysOf(Array_,Array_[:-1]) )
-    return([Array]+SubArraysOf(Array[1:],Array_))
+        return( subArraysOf(Array_,Array_[:-1]) )
+    return([Array]+subArraysOf(Array[1:],Array_))
 
-def permuter( inputs , n ) :
+def permuter( inputs:list , n:int ) -> list :
     # permuter( inputs = ['T2D','NGT','Female','Male'] , n = 2 )
     return( [p[0] for p in zip(itertools.permutations(inputs,n))] )
 
@@ -692,6 +693,79 @@ def qvalues ( p_values_in , pi0 = None ) :
         qs_ = q
     return qs_
 
+
+
+class Qvalues ( object ) :
+    def __init__( self, pvalues:np.array , method:str = "UNIQUE" , pi0:np.array = None ) :
+        from scipy.stats import rankdata
+        self.rankdata = rankdata
+        self.method   : str      = method
+        self.pvalues  : np.array = pvalues
+        self.qvalues  : np.array = None
+        self.qpres    : np.array = None
+        if method == "FDR-BH" :
+            self.qpres = self.qvaluesFDRBH  ( self.pvalues )
+        if method == "QVALS"  :
+            self.qpres = self.qvaluesFDRBH  ( self.pvalues , pi0 )
+        if method == "UNIQUE" :
+            self.qpres = self.qvaluesUNIQUE ( self.pvalues , pi0 )
+
+    def __str__ ( self ) :
+        return ( self.info() )
+
+    def __repr__( self ) :
+        return ( self.info() )
+
+    def help ( self ) :
+        desc__ = "\n\nRANK CORRECTION FOR P-VALUES\nVIABLE METHODS ARE method = FDR-BH , QVALS , UNIQUE\n\n EMPLOYED METHOD: " + self.method
+        return ( desc__ )
+
+    def info ( self ) :
+        desc__ = "\nMETHOD:"+self.method+"\n   q-values       \t     p-values\n"
+        return ( desc__+'\n'.join( [ ' \t '.join(["%10.10e"%z for z in s]) for s in self.qpres ] ) )
+
+    def get ( self ) :
+        return ( self.qpres )
+
+    def qvaluesFDRBH ( self , p_values_in:np.array = None , pi0:np.array = None ) :
+        p_s = p_values_in
+        if p_s is None :
+            p_s = self.pvalues
+        m = int(len(p_s))
+        if pi0 is None :
+            pi0 = np.array([1. for i in range(m)])
+        qs_ = []
+        ps = p_s
+        frp_ = (self.rankdata( ps,method='ordinal' )-0.5)/m
+        ifrp_ = [ ( (p<=f)*f + p*(p>f) ) for p,f in zip(ps,frp_) ]
+        for ip,p0 in zip(range(m),pi0) :
+            p_ = ps[ ip ] ; f_ = frp_[ip]
+            q_ = p0 * p_ / ifrp_[ip]
+            qs_.append( (q_,p_) )
+        self.qvalues = np.array([q[0] for q in qs_])
+        return np.array(qs_)
+
+    def qvaluesUNIQUE ( self , p_values_in = None , pi0 = None ) :
+        p_s = p_values_in
+        if p_s is None :
+            p_s = self.pvalues
+        m = int(len(set(p_s)))
+        n = int(len(p_s))
+        if pi0 is None :
+            pi0 = np.array([1. for i in range(n)])
+        qs_ = []
+        ps  = p_s
+        frp_  = (self.rankdata( ps,method='average' )-0.5)/m
+        ifrp_ = [ ( (p<=f)*f + p*(p>f) ) for p,f in zip(ps,frp_) ]
+        for ip,p0 in zip( range(n),pi0 ) :
+            p_ = ps[ ip ] ; f_ = frp_[ip]
+            q_ = p0 * p_ / ifrp_[ip]
+            qs_.append( (q_,p_) )
+        self.qvalues = np.array([q[0] for q in qs_])
+        return np.array(qs_)
+
+
+
 class MultiFactorAnalysis ( object ) :
     def __init__( self, analyte_df, journal_df, formula ) :
         #super(MultiFactorAnalysis,self).__init__()
@@ -1518,11 +1592,103 @@ def add_kendalltau( analyte_results_df , journal_df , what='M' , sample_names =
         analyte_results_df['KendallTau'] = K
     return ( analyte_results_df )
 
+# TP FP FN TN
+def quality_metrics ( TP:int , FP:int , FN:int , TN:int , alternative:str='two-sided') -> dict :
+    oddsratio , pval = stats.fisher_exact([[TP, FP], [FN, TN]], alternative=alternative )
+    results_lookup = { 'TP':TP , 'TN':TN ,
+                'FN':FN ,'FP':FP ,
+                'causality'   : ( TP+TN+1 ) / ( FP+FN+1 ) ,
+                'sensitivity' : TP / ( TP+FN ) ,
+                'specificity' : TN / ( TN+FP ) ,
+                'precision'   : TP / ( TP+FP ) ,
+                'recall'      : TP / ( TP+FN ) ,
+                'accuracy'    : ( TP+TN ) / ( TP+TN+FP+FN ) ,
+                'negation'    : TN / ( TN+FN ) , # FNR
+                'FPR:'        : FP / ( FP+TN ) , # False positive rate
+                'FDR'         : FP / ( FP+TP ) , # False discovery rate
+                'F1'          : 2 * TP / ( TP + FP + TP + FN ) , # 2 * GEOMMEAN(recall,precision)
+                'MCC'         : (TP*TN-FP*FN) / np.sqrt( (TP+FP)*(TP+FN)*(TN+FP)*(TN+FN) ) , # MATTHEWS CORR COEF
+                'Fishers odds ratio' : oddsratio ,
+                'Fishers p-value'    : pval ,
+                'Odds correct'  : ( TP / FP ) ,
+                'Odds incorrect': ( FN / TN ) ,
+                'PLR'         : ( TP / FP ) * ( FP + TN ) / ( TP + FN ) , # POSITIVE LIKELIHOOD RATIO
+                'NLR'         : ( FN / TN ) * ( FP + TN ) / ( TP + FN )   # NEGATIVE LIKELIHOOD RATIO
+        }
+    return ( results_lookup )
 
-def calculate_rates( journal_df , inferred_df ,
-                     formula , inference_label = 'owner',
-                     bVerbose = False ,
-                     strictness = 'intersect' ) :
+def pvalues_dsdr_n ( v:np.array , bReturnDerivatives:bool=False ) -> np.array :
+    #
+    # PVALUES FROM "RANK DERIVATIVES"
+    #
+    N = len(v)
+
+    def nn ( N:int , i:int , n:int=1 )->list :
+        t = [(i-n)%N,(i+n)%N]
+        if i-n<0 :
+            t[0] = 0
+            t[1] += n-i
+        if i+n>=N :
+            t[0] -= n+i-N
+            t[1] = N-1
+        return ( t )
+
+    import scipy.stats as st
+    rv = st.rankdata(v)-1
+    vr = { int(k):v for k,v in zip(rv,range(len(rv)))}
+    ds = []
+    for w,r in zip(v,rv) :
+        nr = nn(N,int(r),1)
+        nv = [ vr[j] for j in nr]
+        s_ = sorted(list(set([*[v[vr[int(r)]]], *[v[j] for j in nv]])) )
+        ds.append( np.mean( np.diff(s_) )) # DR IS ALWAYS 1
+    M_,Var_ = np.mean(ds) , np.std(ds)**2
+    from scipy.special import erf as erf_
+    loc_Q   = lambda X,mean,variance : [ 1. - 0.5*( 1. + erf_(  (x-mean)/np.sqrt( 2.*variance ) ) ) for x in X ]
+    rv = loc_Q ( ds,M_,Var_ ) # KEEP CONSERVATIVE
+    loc_E  = lambda X,L_mle : [ np.exp(-L_mle*x) for x in X ]
+    ev = loc_E ( ds,1.0/M_)   # ADD FOR REFERENCE
+    if bReturnDerivatives :
+        rv = [*rv,*ds,*ev ]
+    return ( np.array(rv).reshape(-1,N) )
+
+
+def pvalues_dsdr_e ( v:np.array , bReturnDerivatives:bool=False ) -> np.array :
+    #
+    # PVALUES FROM "RANK DERIVATIVES"
+    #
+    N = len(v)
+
+    def nn ( N:int , i:int , n:int=1 )->list :
+        t = [(i-n)%N,(i+n)%N]
+        if i-n<0 :
+            t[0] = 0
+            t[1] += n-i
+        if i+n>=N :
+            t[0] -= n+i-N
+            t[1] = N-1
+        return ( t )
+
+    import scipy.stats as st
+    rv = st.rankdata(v)-1
+    vr = { int(k):v for k,v in zip(rv,range(len(rv)))}
+    ds = []
+    for w,r in zip(v,rv) :
+        nr = nn(N,int(r),1)
+        nv = [ vr[j] for j in nr]
+        s_ = sorted(list(set([*[v[vr[int(r)]]], *[v[j] for j in nv]])) )
+        ds.append( np.mean( np.diff(s_) )) # DR IS ALWAYS 1
+    M_ = np.mean(ds)
+    loc_E  = lambda X,L_mle : [ np.exp(-L_mle*x) for x in X ]
+    ev = loc_E ( ds,1.0/M_)   # EXP DISTRIBUTION P
+    if bReturnDerivatives :
+        rv = [*ev,*ds ]
+    return ( np.array(rv).reshape(-1,N) )
+
+def calculate_rates( journal_df:pd.DataFrame , inferred_df:pd.DataFrame ,
+                     formula:str , inference_label:str = 'owner',
+                     bVerbose:bool = False ,
+                     strictness:str = 'intersect' ) -> dict :
 
     strictness_function = { 'any':any,'intersect':lambda x:x }
 
@@ -1565,18 +1731,8 @@ def calculate_rates( journal_df , inferred_df ,
     TN = np.sum( np.sum( ( not_inferred_OH * not_known_OH ).apply(strictness_function[strictness]) ) )
     FN = np.sum( np.sum( ( not_inferred_OH * known_OH     ).apply(strictness_function[strictness]) ) )
 
-    results_lookup = { 'TP':TP , 'TN':TN ,
-                'FN':FN ,'FP':FP ,
-                'causality'   : ( TP+TN ) / ( FP+FN ) ,
-                'sensitivity' : TP / ( TP+FN ) ,
-                'specificity' : TN / ( TN+FP ) ,
-                'precision'   : TP / ( TP+FP ) ,
-                'recall'      : TP / ( TP+FN ) ,
-                'accuracy'    : ( TP+TN ) / ( TP+TN+FP+FN ) ,
-                'negation'    : TN / ( TN+FN ) , # FNR
-                'FPR:'        : FP / ( FP+TN ) , # False positive rate
-                'FDR'         : FP / ( FP+TP )   # False discovery rate
-    }
+    results_lookup = quality_metrics ( TP , FP , FN , TN )
+
     return ( results_lookup )
 
 
@@ -1608,16 +1764,16 @@ def isItPrime( N , M=None,p=None,lM05=None ) :
 
 # FIRST APPEARENCE:
 # https://gist.github.com/richardtjornhammar/ef1719ab0dc683c69d5a864cb05c5a90
-def Fibonacci(n):
+def fibonacci(n:int) -> int :
     if n-2>0:
-        return ( Fibonacci(n-1)+Fibonacci(n-2) )
+        return ( fibonacci(n-1)+fibonacci(n-2) )
     if n-1>0:
-        return ( Fibonacci(n-1) )
+        return ( fibonacci(n-1) )
     if n>0:
        return ( n )
 
-def F_truth(i):
-    return ( Fibonacci(i)**2+Fibonacci(i+1)**2 == Fibonacci(2*i+1))
+def f_truth(i:int) -> bool : #  THE SQUARE SUM OF THE I:TH AND I+1:TH FIBONACCI NUMBER ARE EQUAL TO THE FIBONACCI NUMBER AT POSITION 2I+1
+    return ( fibonacci(i)**2+fibonacci(i+1)**2 == fibonacci(2*i+1))
 
 if __name__ == '__main__' :
 
